# frozen_string_literal: true
require 'listen'
require 'observer'
require 'pathname'

module FileIndexer

  # A simple file indexer, keeps a database of files and associated metadata
  class Indexer
    include Observable

    # The directories that will be indexed
    attr_accessor :dirs
    # Only files with these extensions will be considered.  Just use '*' for all
    # files.
    attr_accessor :exts
    # This is a Proc that will be called for every file and its result will be
    # stored in the database
    attr_accessor :action
    # This is the actual database of files, it is a Hash
    attr_accessor :files
    # This is the filename which will be used as the persistent database
    attr_reader :db_file

    def initialize(dirs, exts, db_file, &action)
      @dirs    = dirs.map { |d| sanitize_dirpath(d) }
      @exts    = exts
      @db_file = db_file
      @action  = action

      @files    = {}
      @listener = nil

      @files.merge!(read_db)
      prune!
      index_all!
    end

    # Index multiple directories
    def index_all(dirs = @dirs, exts = @exts)
      dirs.map { |dir| index(glob(dir, exts)) }.reduce(&:merge)
    end

    # @see #index_all
    def index_all!(dirs = @dirs, exts = @exts)
      files.merge!(index_all(dirs, exts))
      changed!
      files
    end

    # Index a list of files
    def index(fs)
      fs.map do |file|
        next if files.key?(file)
        [file, action.(file)]
      end.compact.to_h
    end

    # @see #index
    def index!(fs)
      files.merge!(index(fs))
      changed!
      files
    end

    # Prune nonexistant files from the index
    def prune!
      files.select! { |f,_| File.exist?(f) }.to_h
    end

    # Watch directories for changes using the listen gem, when called make sure
    # to call #stop when done watching.
    def watch(dirs = @dirs, exts = @exts)
      @listener =
        Listen.to(*dirs, only: /\.(#{exts.join('|')})$/) do |mod, add, del|
          del.each { |f| files.delete(f) }
          index!(mod + add)
        end
      @listener.start
    end

    # Stop watching directories for changes
    def stop
      @listener&.stop
    end

    # Write the databse to disk
    def write_db
      File.open(db_file, 'w') { |f| f.puts(Marshal.dump(files)) }
    end

    # Read the database from disk
    def read_db
      return {} unless File.exist?(db_file)
      File.open(db_file, 'r') { |f| Marshal.load(f.read) }
    end

    private

    # Makes sure a directory exists, and sanitizes it
    def sanitize_dirpath(dir)
      Pathname.new(File.expand_path(dir)).realpath.to_s
    end

    # Glob all files with certain extensions in a directory
    def glob(dir, exts = @exts)
      Dir.glob("#{dir}**/*.{#{exts.join(',')}}")
    end

    # Notify observers of changes and rewrite the database file
    def changed!(*args)
      changed
      write_db
      notify_observers(*args)
    end
  end
end
